# -*- coding: utf-8 -*-
# TinyOlap, copyright (c) 2021 Thomas Zeutschler
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import itertools
import sqlparse

from database import Database
from cube import Cube


class Query:
    """
    Basic implementation of a SQL query interface for TinyOlap cubes.
    """

    def __init__(self, db: Database, sql: str = None,
                 include_column_names: bool = False):
        """
        Create a new query object for the given database.
        :param db: The database to query against.
        :param sql: (optional) An SL statement to be executed.
        :param include_column_names: Include column names in the resultset.
        :param include_cube_name: Include a cube name column in the resultset.
        """
        self.database = db
        self.sql = sql
        self._records = []
        self._include_column_names = include_column_names

    @property
    def include_column_names(self):
        return self._include_column_names

    @include_column_names.setter
    def include_column_names(self, value: bool):
        self._include_column_names = value

    @property
    def records(self):
        """
        The records returned by the query. Records contain the address and the current value of the queried cells.
        :return: An array of arrays, the outer array represents the rows
        """
        return self._records

    def execute(self, sql: str = None):
        """
        Executes an SQL statement against the defined database.
        :param sql: The SQL statement to execute.
        :return: True, if the execution was successful, False otherwise.
        :raises NotImplementedError: Raised if a certain SQL language capability is not supported.
        :raises KeyError: If a certain name (e.g. cube, dimension, member, subset) does not exists in the database.
        :raises SyntaxError: If a SQL statement contains a syntax error or is incomplete.
        """
        if sql:
            self.sql = sql
        self._records = []
        statement = sqlparse.split(self.sql)
        if len(statement) > 1:
            raise NotImplementedError(f"Multiple sql statements are not yet supported. "
                                      f"A single sql statement was expected, but multiple satements were "
                                      f"found in {self.sql}.")
        self.parsed = sqlparse.parse(statement[0])[0]

        tokens = self.__remove_whitespace_tokens(self.parsed.tokens)
        cube: Cube
        cube_name = None

        # resolve the cube. The token after the 'FROM' keyword should contain the cube name
        from_index = -1
        for index, token in enumerate(tokens):
            if token.is_keyword and token.normalized == "FROM":
                if index == len(tokens) - 1:  # FROM token is already the last token.
                    raise KeyError("Cube name missing after <FROM> clause.")
                from_index = index
                cube_name = tokens[index + 1].normalized
                # check for cube name
                cube = self.database.cubes.lookuptry(cube_name)
                if not cube:
                    raise KeyError(f"Cube '{cube_name}' does not exit in database '{self.database.name}'.")
                cube_name = cube.name
                break
        else:
            raise SyntaxError("<FROM> clause missing in statement.")

        # validate and prepare the query
        where_slicer = self.__resolve_where_clause(tokens)
        select_slicer = self.__resolve_select_clause(from_index, tokens)
        query_def = self.__validate_query(cube, select_slicer, where_slicer)

        # execute the query
        records = []
        if self._include_column_names:
            records.append(self.__get_column_names(query_def))
        members_tuples = [tuple(dimension["members"]) for dimension in query_def["dims"]]
        addresses = itertools.product(*members_tuples)
        if not select_slicer:
            for address in addresses:
                value = cube[address]  # get a value from the database
                record = list(address)
                record.append(value)
                records.append(record)
        else:
            fields = query_def["fields"]
            for address in addresses:
                record = []
                value_already_appended = False
                value = None
                for field in fields:
                    value = cube[address]  # get a value from the database
                    if type(field) is int:
                        if field == -1:
                            record.append(value)
                            value_already_appended = True
                        else:
                            record.append(address[field])
                    else:
                        # get an attribute for the current member of a dimension
                        member = address[field["member_index"]]
                        attribute = field["dimension"].get_attribute(field["attribute"], member)
                        record.append(attribute)

                if not value_already_appended:
                    record.append(value)

                records.append(record)


            pass

        self._records = records
        return True

    def __get_column_names(self, query_def):
        col_names = []
        value_already_appended = False
        if "fields" in query_def:
            for field in query_def["fields"]:
                if field == -1:
                    col_names.append("value")
                    value_already_appended = True
                elif type(field) is dict:
                    col_names.append(f"{field['dimension'].name}.{field['attribute']}")
                else:
                    col_names.append(query_def["dims"][field]["dimension"].name)
        else:
            for dimension in query_def["dims"]:
                col_names.append(dimension["dimension"].name)
        if not value_already_appended:
            col_names.append("value")
        return col_names

    def __resolve_where_clause(self, tokens):
        where_slicer = []
        where_tokens = [token for token in tokens if type(token) is sqlparse.sql.Where]
        if where_tokens:
            if type(where_tokens[0].tokens[-1]) is sqlparse.sql.IdentifierList:
                identifiers = where_tokens[0].tokens[-1].get_identifiers()
                where_slicer = [self.__strip(str(x)) for x in identifiers]
        return where_slicer

    def __resolve_select_clause(self, from_index, tokens):
        select_tokens = tokens[1:from_index]
        if len(select_tokens) == 1 and select_tokens[0].normalized =="*":
            return []  # special case: SELECT * FROM ...
        select_slicer = self.__resolve_identifiers(
            [token for token in select_tokens[0].tokens])
        return select_slicer

    def __validate_query(self, cube: Cube, select_slicer: list, where_slicer: list):
        dim_count = cube.dimensions_count
        db = cube._database
        query_def = {"cube": cube, "db": db, "dim_count": dim_count,
                     "dims": [{"dimension": cube.get_dimension_by_index(d),
                               "members": [],
                               "expression": None} for d in range(cube.dimensions_count)]}

        # Process the WHERE statement
        unresolved_dims = list(range(dim_count))
        for slice in where_slicer:
            dim, member = self.__dim_member_split(slice)
            if dim:  # a dimension name is defined, e.g.: years:'2022'
                for index, dimension in enumerate(query_def["dims"]):
                    if dimension["dimension"].name.lower() == dim.lower():
                        if dimension["dimension"].member_exists(member):
                            dimension["members"].append(member)
                            unresolved_dims.remove(index)
                        else:
                            # check for subset name
                            if member in dimension["dimension"].subsets:
                                dimension["members"] = dimension["dimension"].subsets[member][3]
                                unresolved_dims.remove(index)
                            if member == "*":
                                dimension["members"] = dimension["dimension"].get_members()
                                unresolved_dims.remove(index)

                            # check for member list e.g.: (Jan, 'Feb')
                            elif member.startswith('(') and member.endswith(')'):
                                members = self.__strip_brackets(member)
                                members = members.split(",")
                                members = [self.__strip(member) for member in members]
                                invalid_members = []
                                for member in members:
                                    if not dimension["dimension"].member_exists(member):
                                        invalid_members.append(member)
                                if invalid_members:
                                    raise KeyError(f"Invalid member list for dimension {dimension['dimension'].name} "
                                                   f" in WHERE statement found. The following members do not exist: "
                                                   f"{', '.join(invalid_members)}")
                                dimension["members"] = members
                                unresolved_dims.remove(index)

            else:  # only a member name is defined, e.g.: '2022' > resolve the dimension
                found_index = -1
                for d in unresolved_dims:
                    if query_def["dims"][d]["dimension"].member_exists(member):
                        query_def["dims"][d]["members"].append(member)
                        found_index = d
                        break
                if found_index > -1:
                    unresolved_dims.remove(found_index)
                else:
                    # If it's not a member then it might be a list of members or the name of a subset.

                    raise KeyError(f"Unresolvable member '{slice}' in WHERE statement found.")

        # for all unresolved dimensions, get the (default) first member of the dimension.
        for d in unresolved_dims:
            member = query_def["dims"][d]["dimension"].get_first_member()
            query_def["dims"][d]["members"].append(member)

        # If there is no specific SELECT statement, we're already done
        #   esp. for 'SELECT * FROM ...' statements
        query_def["is_cell_request"] = True
        if not select_slicer:
            return query_def

        # Process the SELECT statement
        fields = []
        for slicer in select_slicer:
            found = False

            # search for 'value' keyword
            if slicer.lower() == "value":
                fields.append(-1)
                found = True

            # search for dimension names, if found remember their index
            if not found:
                for d in range(dim_count):
                    if query_def["dims"][d]["dimension"].name.lower() == slicer.lower():
                        found = True
                        fields.append(d)
                        break

            # search for attribute names
            if not found:
                dim, attribute = self.__dim_attribute_split(slicer)
                found = attribute is not None
                if found:
                    found = False
                    for d in range(dim_count):
                        if query_def["dims"][d]["dimension"].name.lower() == dim.lower():
                            if query_def["dims"][d]["dimension"].has_attribute(attribute):
                                fields.append({"dimension":query_def["dims"][d]["dimension"],
                                               "attribute": attribute,
                                               "member_index": d})
                                found = True
                                break
                    # if not found:
                    #     raise KeyError(f"Unresolvable attribute name '{slice}' "
                    #                    f"after SELECT statement found.")

            # not supported or unable to resolve
            if not found:
                raise KeyError(f"Unresolvable field or keyword '{slice}' "
                               f"after SELECT statement found.")

        query_def["fields"] = fields
        return query_def

    def __dim_member_split(self, expression: str):
        dim = None
        member = None
        if "=" in expression:
            tokens = expression.split("=", maxsplit=1)
            dim = self.__strip(tokens[0])
            member = self.__strip(tokens[1])
        elif ":" in expression:
            tokens = expression.split(":", maxsplit=1)
            dim = self.__strip(tokens[0])
            member = self.__strip(tokens[1])
        else:
            member = expression
        return dim, member

    def __dim_attribute_split(self, expression: str):
        dim = None
        attribute = None
        if "." in expression:
            tokens = expression.split(".", maxsplit=1)
            dim = self.__strip(tokens[0])
            attribute = self.__strip(tokens[1])
        elif ":" in expression:
            tokens = expression.split(":", maxsplit=1)
            dim = tokens[0]
            attribute = self.__strip(tokens[1])
        else:
            dim = expression
        return dim, attribute

    @staticmethod
    def __strip(text: str):
        """
        Removes surrounding double or single quotes from a text. Even if nested.
        :param text: The text to be stripped.
        :return: The stripped version of the text.
        """
        text = text.strip()
        while True:
            if text.startswith("'") and text.endswith("'") and len(text) > 1:
                text = text[1:-1].strip()
            elif text.startswith('"') and text.endswith('"') and len(text) > 1:
                text = text[1:-1].strip()
            else:
                return text

    @staticmethod
    def __strip_brackets(text: str):
        """
        Removes surrounding brackets [], {} or (). Even if nested.
        :param text: The text to be stripped.
        :return: The stripped version of the text.
        """
        text = text.strip()
        while True:
            if text.startswith("(") and text.endswith(")") and len(text) > 1:
                text = text[1:-1].strip()
            elif text.startswith("[") and text.endswith("]") and len(text) > 1:
                text = text[1:-1].strip()
            elif text.startswith("{") and text.endswith("}") and len(text) > 1:
                text = text[1:-1].strip()
            else:
                return text

    @staticmethod
    def __remove_whitespace_tokens(tokens):
        return [token for token in tokens if not token.is_whitespace]

    def __resolve_identifiers(self, tokens):
        return [self.__strip(str(token.normalized)) for token in tokens if type(token) is sqlparse.sql.Identifier]

    def __str__(self):
        return self.sql

    def __repr__(self):
        return self.sql



from __future__ import annotations
from abc import ABC, abstractmethod
import datetime
from tinyolap.storage.storageprovider import StorageProvider


class MockStorage(StorageProvider):
    """Storage provider mock up. Does not store anything, just for testing purposes."""

    # region basic database handling - open , close, delete, ...
    @property
    def connected(self) -> bool:
        """
        Identifies if the storage provider is connected or open.
        :return: ``True`` if the storage provider is connected, ``False`` otherwise.
        """
        return True

    @property
    def uri(self) -> str:
        """
        Returns the uri (uniform resource identifier) of the current database.
        :return: The uri of the databse.
        """
        return ""

    def open(self, **kwargs) -> bool:
        """
        Opens the database. If the database does not exist, a new database file will be created.
        :param kwargs:
        :return: Returns ``True``if the database has been opened successfully.
        :raises DatabaseBackendException: Raised when the opening of / connecting to the database file failed.
        """
        return True

    def close(self):
        """
        Closes the database.
        :return: ``True`` if the database was successfully closed
            or was already closed, ``False``otherwise.
        """
        pass

    def exists(self) -> bool:
        """
        Checks if a database with the given name exists.
        :return: ``True`` if the database file exists, ``False``otherwise.
        """
        return True

    def delete(self):
        """
        Deletes the database file and the database log file, if such exist.
        :return: True if the database file and the database log file was deleted successfully
        or if both do not exist.
        """
        pass

    def delete_log(self):
        """
        Deletes the database log file, if such exists.
        :return: True if the log file was deleted successfully or if the log file not exist.
        """
        pass
    # endregion

    # region History related methods
    def add_to_history(self, time_stamp: datetime.datetime, user: str, action: str, data: str):
        """
        Add an event to the database history.
        :param time_stamp: Timestamp of the event
        :param user: The name of the user that initiated the event.
        :param action: The name of action executed through the event.
        :param data: The data related to the event. Defines what and how to undo or redo the event.
        """
        pass

    def add_many_to_history(self, records: list[tuple[datetime.datetime, str, str, str]]):
        """
        Add multiple events to the database history.
        :param records: A list of tuples of format (time_stamp:datetime, user:str, action:str, data:str)
        """
        pass

    def clear_history(self):
        """
        Clears the entire history.
        """
        pass

    def delete_history_earlier_than(self, time_stamp: datetime.datetime):
        """
        Deletes all events from the history earlier than a specific timestamp.
        :param time_stamp: The timestamp to apply.
        """
        pass

    def delete_history_later_than(self, time_stamp: datetime.datetime):
        """
        Deletes all events from the history later than a specific timestamp.
        :param time_stamp: The timestamp to apply.
        """
        pass

    def delete_history_by_rowid(self, rowid):
        """
        Deletes events from the history by their rowid. Argument ``rowids`` can either be a single
        int value or an iterable of int.
        :param rowid: Either a single int value, or an iterable of int.
        """
        pass

    def get_history_from_time_window(self, from_time_stamp: datetime.datetime = datetime.datetime.min,
                                     to_time_stamp: datetime.datetime = datetime.datetime.max)\
            -> list[tuple[datetime.datetime, str, str, str, int]]:
        """
        Returns the history records for a given time window.
        :param from_time_stamp: Begin of the time window.
        :param to_time_stamp: End of the time window.
        :return: A list of tuples of format (time_stamp:datetime, user:str, action:str, data:str, rowid:int)
        sorted in ascending order by the timestamp.
        """
        return []

    def get_history(self, count: int = 1) -> list[tuple[datetime.datetime, str, str, str, int]]:
        """
        Returns the a certain number of records from the history.
        :param count: Number of records to be returned.
        :return: A list of tuples of format (time_stamp:datetime, user:str, action:str, data:str, rowid:int)
        sorted in ascending order by the timestamp.
        """
        return []

    def count_history(self) -> int:
        """
        Returns the number of records in the history.
        """
        return 0
    # endregion

    # region Cell (data record) related methods
    def set_record(self, cube_name, address: str, data: str = None, instant_commit: bool = True):
        """
        Add, updates or deletes a record related to a specific address and cube.
        To delete data, set argument ``data`` to ``None``.
        :param cube_name: Name of the targeted cube.
        :param address: The address of the cube cell as a string.
          Recommendation: use *str(idx_address:tuple[int])* to generate address from int cell address tuples.
        :param data: The data to be set as a string. The json format should be applied, or any string.
        :param instant_commit: If set to ``True``, then the change is instantly committed to the underlying database.
        :return: ``True`` if successful.
        """
        pass

    def set_records(self, cube_name, records: list[tuple[str, str]], instant_commit: bool = True):
        """
        Add, updates or deletes a record related to a specific address and cube.
        To delete data, set argument ``data`` to ``None``.
        :param cube_name: Name of the targeted cube.
        :param records: A list of (address, data) tuples to be set for the specific cube.
        :param instant_commit: If set to ``True``, then the change is instantly committed to the underlying database.
        :return: ``True`` if successful.
        """
        pass

    def get_record(self, cube_name: str, address: str) -> str:
        """
        Returns data from a cubes data table.
        :param cube_name: Name of the cube to get data from.
        :param address: The requested address.
        :return: The data stored for the given address.
           If the address does not exist, ``None`` will be returned.
        """
        return None

    def get_records(self, cube_name: str):
        """
        Returns all records from a cubes data table.
        :param cube_name: Name of the cube to get data from.
        :return: All records of the cube.
        """
        pass
    # endregion

    # region meta data related methods
    @abstractmethod
    def add_meta(self, key: str, json: str):
        """
        Adds or updates a meta configuration.
        :param key: The name of the meta item.
        :param json: The configuration of the meta item in json format.
        """
        pass

    @abstractmethod
    def get_meta(self, key: str) -> str:
        """
        Returns the configuration of a meta item in json format.
        :return: A json string.
        """
        return None
    # endregion

    # region Cube related methods
    def get_cubes(self) -> list[tuple[str, str]]:
        """
        Returns a list of all cubes and their configuration (in json format)
        available in the database.
        :return: A list of tuples of type (cube_name:str, json:str).
        """
        return []

    def get_cube_names(self) -> list[str]:
        """
        Returns a list of all cubes names available in the database.
        :return: List of cube names.
        """
        return []

    def count_cubes(self) -> int:
        """
        Returns the number of cubes defined in the database.
        :return: The number of cubes defined for the database.
        """
        return 0

    def add_cube(self, cube_name: str, json: str):
        """
        Adds or updates a cube configuration.
        :param dimensions_count: Number of the dimension in the cube.
          If set to -1, then no data table will be created.
        :param cube_name: The name of the cube.
        :param json: The configuration of the cube in json format.
        :return: ``True`` if successful,``False`` otherwise.
        """
        pass

    def remove_cube(self, cube_name: str):
        """
        Removes a cube from the database.
        :param cube_name: Name of the cube to be removed.
        """
        pass

    def clear_cube(self, cube_name: str):
        """
        Clears (deletes) all records from a cubes data table.
        :param cube_name: The name of the cube to be cleared.
        :return:
        """
        pass

    def count_cube_records(self, cube_name: str) -> int:
        """
        Returns the number of records contained in cube table.
        :return: The number of records in cube table.
        """
        return 0

    # endregion

    # region Dimension related methods
    def get_dimensions(self) -> list[tuple[str, str]]:
        """
        Returns a list of all dimensions and their configuration (in json format)
        available in the database.
        :return: A list of tuples of type (cube_name:str, json:str).
        """
        return []

    def get_dimension_names(self) -> list[str]:
        """
        Returns a list of all dimension names available in the database.
        :return: List of dimension names.
        """
        return []

    def count_dimensions(self) -> int:
        """
        Returns the number of dimensions defined in the database.
        :return: The number of dimension defined for the database.
        """
        return 0

    def add_dimension(self, dimension_name: str, json: str):
        """
        Adds or updates a dimension configuration.
        :param dimension_name: The name of the dimension.
        :param json: The configuration of the dimension in json format.
        :return: ``True`` if successful,``False`` otherwise.
        """
        pass

    def remove_dimension(self, dimension_name: str):
        """
        Removes a dimension from the database.
        :param dimension_name: Name of the dimension to be removed.
        """
        pass
    # endregion




